/*
 * playbackbackend.cpp
 *
 *  Created on: 05.12.2010
 *      Author: darkstar
 */

#include "playbackbackend.h"

static PlaybackBackendRegistry instance;

PlaybackBackendRegistry &PlaybackBackendRegistry::singleton()
{
    return instance;
}

AbstractPlaybackBackend *PlaybackBackendRegistry::createInstanceOfDefaultBackend()
{
    return defaultFactory()->createInstance();
}

AbstractPlaybackBackend *PlaybackBackendRegistry::createNewInstance(const QString &className)
{
    AbstractPlaybackBackendFactory *factory = findFactory(className);
    if (factory)
        return factory->createInstance();
    else
        return NULL;
}

void PlaybackBackendRegistry::registerFactory(AbstractPlaybackBackendFactory *factory)
{
    if (!findFactory(factory->internalClassName()))
    {
        factories_.append(factory);
        factoriesMap_.insert(factory->internalClassName(), factory);
    }
}

AbstractPlaybackBackendFactory *PlaybackBackendRegistry::defaultFactory()
{
    return factories_.at(0);
}

AbstractPlaybackBackendFactory *PlaybackBackendRegistry::findFactory(const QString &className)
{
#ifdef QT4
    QHash<QString, AbstractPlaybackBackendFactory *>::const_iterator it = factoriesMap_.find(className);

    if (it != factoriesMap_.end())
        return it.value();
    else
        return NULL;
#else
    return factoriesMap_.find(className);
#endif
}
